# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Alexey Anisenkov, anisyonk@cern.ch, 2018

"""
Job specific Info Service
It could customize/overwrite settings provided by the main Info Service

:author: Alexey Anisenkov
:contact: anisyonk@cern.ch
:date: January 2018
"""

from .infoservice import InfoService
from .jobinfo import JobInfoProvider

import logging
logger = logging.getLogger(__name__)


class JobInfoService(InfoService):  ## TO BE DEPRECATED/REMOVED
    """
        Info service: Job specific
        Job could overwrite settings provided by Info Service

        *** KEPT for a while in repo .. most probably will be deprecated and removed soon **
    """

    def __init__(self, job):

        self.jobinfo = JobInfoProvider(job)



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Wen Guan, wen.guan@cern.ch, 2018
# - Paul Nilsson, paul.nilsson@cern.ch, 2019


import time

from pilot.common import exception
from pilot.common.pluginfactory import PluginFactory

import logging
logger = logging.getLogger(__name__)

"""
Main class to manage the event service work.
"""


class WorkExecutor(PluginFactory):

    def __init__(self, args=None):
        super(WorkExecutor, self).__init__()
        self.payload = None
        self.plugin = None
        self.is_retrieve_payload = False
        self.args = args
        self.pid = None

    def get_pid(self):
        return self.plugin.get_pid() if self.plugin else None

    def set_payload(self, payload):
        self.payload = payload

    def set_retrieve_paylaod(self):
        self.is_retrieve_payload = True

    def get_payload(self):
        return self.payload

    def get_plugin_confs(self):
        plugin_confs = {}
        if self.args and 'executor_type' in list(self.args.keys()):  # Python 2/3
            if self.args['executor_type'] == 'hpo':
                plugin_confs = {'class': 'pilot.eventservice.workexecutor.plugins.hpoexecutor.HPOExecutor'}
            elif self.args['executor_type'] == 'raythena':
                plugin_confs = {'class': 'pilot.eventservice.workexecutor.plugins.raythenaexecutor.RaythenaExecutor'}
            elif self.args['executor_type'] == 'generic':
                plugin_confs = {'class': 'pilot.eventservice.workexecutor.plugins.genericexecutor.GenericExecutor'}
            elif self.args['executor_type'] == 'base':
                plugin_confs = {'class': 'pilot.eventservice.workexecutor.plugins.baseexecutor.BaseExecutor'}
            elif self.args['executor_type'] == 'nl':  # network-less
                plugin_confs = {'class': 'pilot.eventservice.workexecutor.plugins.nlexecutor.NLExecutor'}
            elif self.args['executor_type'] == 'boinc':
                plugin_confs = {'class': 'pilot.eventservice.workexecutor.plugins.boincexecutor.BOINCExecutor'}
            elif self.args['executor_type'] == 'hammercloud':  # hammercloud test: refine normal simul to ES
                plugin_confs = {'class': 'pilot.eventservice.workexecutor.plugins.hammercloudexecutor.HammerCloudExecutor'}
            elif self.args['executor_type'] == 'mpi':  # network-less
                plugin_confs = {'class': 'pilot.eventservice.workexecutor.plugins.mpiexecutor.MPIExecutor'}
        else:
            plugin_confs = {'class': 'pilot.eventservice.workexecutor.plugins.genericexecutor.GenericExecutor'}

        plugin_confs['args'] = self.args
        return plugin_confs

    def start(self):
        plugin_confs = self.get_plugin_confs()
        logger.info("Plugin confs: %s" % plugin_confs)
        self.plugin = self.get_plugin(plugin_confs)
        logger.info("WorkExecutor started with plugin: %s" % self.plugin)
        if not self.plugin:
            raise exception.SetupFailure("No available executor plugin.")

        if self.is_retrieve_payload:
            self.payload = self.plugin.set_retrieve_payload()
        else:
            if not self.get_payload():
                raise exception.SetupFailure("Payload is not assigned.")
            else:
                self.plugin.set_payload(self.get_payload())

        logger.info("Starting plugin: %s" % self.plugin)
        self.plugin.start()
        logger.info("Waiting for payload to start")
        while self.plugin.is_alive():
            if self.plugin.is_payload_started():
                logger.info("Payload started with pid: %s" % self.get_pid())
                break
            time.sleep(1)

    def stop(self):
        if not self.plugin:
            raise exception.SetupFailure("No available executor plugin.")
        return self.plugin.stop()

    def is_alive(self):
        if not self.plugin:
            raise exception.SetupFailure("No available executor plugin.")
        return self.plugin.is_alive()

    def get_exit_code(self):
        if not self.plugin:
            raise exception.SetupFailure("No available executor plugin.")
        return self.plugin.get_exit_code()

    def get_event_ranges(self):
        if not self.plugin:
            raise exception.SetupFailure("No available executor plugin.")
        return self.plugin.get_event_ranges()

    def update_events(self, messages):
        if not self.plugin:
            raise exception.SetupFailure("No available executor plugin.")
        return self.plugin.update_events(messages)



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Paul Nilsson, paul.nilsson@cern.ch, 2019

import os

# from pilot.util.container import execute
from pilot.common.errorcodes import ErrorCodes

import logging
logger = logging.getLogger(__name__)

errors = ErrorCodes()


def verify_setup_command(cmd):
    """
    Verify the setup command.

    :param cmd: command string to be verified (string).
    :return: pilot error code (int), diagnostics (string).
    """

    ec = 0
    diagnostics = ""

    return ec, diagnostics


def get_setup_command(job, prepareasetup):
    """
    Return the path to asetup command, the asetup command itself and add the options (if desired).
    If prepareasetup is False, the function will only return the path to the asetup script. It is then assumed
    to be part of the job parameters.

    Handle the case where environmental variables are set -
    HARVESTER_CONTAINER_RELEASE_SETUP_FILE, HARVESTER_LD_LIBRARY_PATH, HARVESTER_PYTHONPATH
    This will create the string need for the pilot to execute to setup the environment.

    :param job: job object.
    :param prepareasetup: not used.
    :return: setup command (string).
    """

    cmd = ""

    # return immediately if there is no release or if user containers are used
    if job.swrelease == 'NULL' or '--containerImage' in job.jobparams:
        logger.debug('get_setup_command return value: {0}'.format(str(cmd)))
        return cmd

    # test if environmental variable HARVESTER_CONTAINER_RELEASE_SETUP_FILE is defined
    setupfile = os.environ.get('HARVESTER_CONTAINER_RELEASE_SETUP_FILE', '')
    if setupfile != "":
        cmd = "source {};".format(setupfile)
        # test if HARVESTER_LD_LIBRARY_PATH is defined
        if os.environ.get('HARVESTER_LD_LIBRARY_PATH', '') != "":
            cmd += "export LD_LIBRARY_PATH=$HARVESTER_LD_LIBRARY_PATH:$LD_LIBRARY_PATH;"
        # test if HARVESTER_PYTHONPATH is defined
        if os.environ.get('HARVESTER_PYTHONPATH', '') != "":
            cmd += "export PYTHONPATH=$HARVESTER_PYTHONPATH:$PYTHONPATH;"
        #unset FRONTIER_SERVER variable
        cmd += "unset FRONTIER_SERVER"

        logger.debug('get_setup_command return value: {0}'.format(str(cmd)))

    return cmd



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Paul Nilsson, paul.nilsson@cern.ch, 2018-2019

# This module contains implementations of job monitoring tasks

import os
import time
from subprocess import PIPE
from glob import glob

from pilot.common.errorcodes import ErrorCodes
from pilot.util.auxiliary import set_pilot_state, show_memory_usage
from pilot.util.config import config
from pilot.util.container import execute
from pilot.util.filehandling import get_directory_size, remove_files, get_local_file_size
from pilot.util.loopingjob import looping_job
from pilot.util.math import convert_mb_to_b, human2bytes
from pilot.util.parameters import convert_to_int, get_maximum_input_sizes
from pilot.util.processes import get_current_cpu_consumption_time, kill_processes, get_number_of_child_processes
from pilot.util.workernode import get_local_disk_space, check_hz

import logging
logger = logging.getLogger(__name__)

errors = ErrorCodes()


def job_monitor_tasks(job, mt, args):
    """
    Perform the tasks for the job monitoring.
    The function is called once a minute. Individual checks will be performed at any desired time interval (>= 1
    minute).

    :param job: job object.
    :param mt: `MonitoringTime` object.
    :param args: Pilot arguments (e.g. containing queue name, queuedata dictionary, etc).
    :return: exit code (int), diagnostics (string).
    """

    exit_code = 0
    diagnostics = ""

    current_time = int(time.time())

    # update timing info for running jobs (to avoid an update after the job has finished)
    if job.state == 'running':
        # confirm that the worker node has a proper SC_CLK_TCK (problems seen on MPPMU)
        check_hz()
        try:
            cpuconsumptiontime = get_current_cpu_consumption_time(job.pid)
        except Exception as e:
            diagnostics = "Exception caught: %s" % e
            logger.warning(diagnostics)
            import traceback
            logger.warning(traceback.format_exc())
            if "Resource temporarily unavailable" in diagnostics:
                exit_code = errors.RESOURCEUNAVAILABLE
            elif "No such file or directory" in diagnostics:
                exit_code = errors.STATFILEPROBLEM
            elif "No such process" in diagnostics:
                exit_code = errors.NOSUCHPROCESS
            else:
                exit_code = errors.GENERALCPUCALCPROBLEM
            return exit_code, diagnostics
        else:
            job.cpuconsumptiontime = int(round(cpuconsumptiontime))
            job.cpuconsumptionunit = "s"
            job.cpuconversionfactor = 1.0
            logger.info('CPU consumption time for pid=%d: %f (rounded to %d)' % (job.pid, cpuconsumptiontime, job.cpuconsumptiontime))

        # check how many cores the payload is using
        set_number_used_cores(job)

        # check memory usage (optional) for jobs in running state
        exit_code, diagnostics = verify_memory_usage(current_time, mt, job)
        if exit_code != 0:
            return exit_code, diagnostics

    # is it time to verify the pilot running time?
#    exit_code, diagnostics = verify_pilot_running_time(current_time, mt, job)
#    if exit_code != 0:
#        return exit_code, diagnostics

    # should the proxy be verified?
    if args.verify_proxy:
        exit_code, diagnostics = verify_user_proxy(current_time, mt)
        if exit_code != 0:
            return exit_code, diagnostics

    # is it time to check for looping jobs?
    exit_code, diagnostics = verify_looping_job(current_time, mt, job)
    if exit_code != 0:
        return exit_code, diagnostics

    # is the job using too much space?
    exit_code, diagnostics = verify_disk_usage(current_time, mt, job)
    if exit_code != 0:
        return exit_code, diagnostics

    # is it time to verify the number of running processes?
    if job.pid:
        exit_code, diagnostics = verify_running_processes(current_time, mt, job.pid)
        if exit_code != 0:
            return exit_code, diagnostics

    # make sure that any utility commands are still running
    if job.utilities != {}:
        utility_monitor(job)

    return exit_code, diagnostics


def set_number_used_cores(job):
    """
    Set the number of cores used by the payload.
    The number of actual used cores is reported with job metrics (if set).

    :param job: job object.
    :return:
    """

    pilot_user = os.environ.get('PILOT_USER', 'generic').lower()
    cpu = __import__('pilot.user.%s.cpu' % pilot_user, globals(), locals(), [pilot_user], 0)  # Python 2/3
    cpu.set_core_counts(job)


def verify_memory_usage(current_time, mt, job):
    """
    Verify the memory usage (optional).
    Note: this function relies on a stand-alone memory monitor tool that may be executed by the Pilot.

    :param current_time: current time at the start of the monitoring loop (int).
    :param mt: measured time object.
    :param job: job object.
    :return: exit code (int), error diagnostics (string).
    """

    show_memory_usage()

    pilot_user = os.environ.get('PILOT_USER', 'generic').lower()
    memory = __import__('pilot.user.%s.memory' % pilot_user, globals(), locals(), [pilot_user], 0)  # Python 2/3

    if not memory.allow_memory_usage_verifications():
        return 0, ""

    # is it time to verify the memory usage?
    memory_verification_time = convert_to_int(config.Pilot.memory_usage_verification_time, default=60)
    if current_time - mt.get('ct_memory') > memory_verification_time:
        # is the used memory within the allowed limit?
        try:
            exit_code, diagnostics = memory.memory_usage(job)
        except Exception as e:
            logger.warning('caught exception: %s' % e)
            exit_code = -1
        if exit_code != 0:
            logger.warning('ignoring failure to parse memory monitor output')
            #return exit_code, diagnostics
        else:
            # update the ct_proxy with the current time
            mt.update('ct_memory')

    return 0, ""


def verify_user_proxy(current_time, mt):
    """
    Verify the user proxy.
    This function is called by the job_monitor_tasks() function.

    :param current_time: current time at the start of the monitoring loop (int).
    :param mt: measured time object.
    :return: exit code (int), error diagnostics (string).
    """

    pilot_user = os.environ.get('PILOT_USER', 'generic').lower()
    userproxy = __import__('pilot.user.%s.proxy' % pilot_user, globals(), locals(), [pilot_user], 0)  # Python 2/3

    # is it time to verify the proxy?
    proxy_verification_time = convert_to_int(config.Pilot.proxy_verification_time, default=600)
    if current_time - mt.get('ct_proxy') > proxy_verification_time:
        # is the proxy still valid?
        exit_code, diagnostics = userproxy.verify_proxy()
        if exit_code != 0:
            return exit_code, diagnostics
        else:
            # update the ct_proxy with the current time
            mt.update('ct_proxy')

    return 0, ""


def verify_looping_job(current_time, mt, job):
    """
    Verify that the job is not looping.

    :param current_time: current time at the start of the monitoring loop (int).
    :param mt: measured time object.
    :param job: job object.
    :return: exit code (int), error diagnostics (string).
    """

    looping_verification_time = convert_to_int(config.Pilot.looping_verification_time, default=600)
    if current_time - mt.get('ct_looping') > looping_verification_time:
        # is the job looping?
        try:
            exit_code, diagnostics = looping_job(job, mt)
        except Exception as e:
            diagnostics = 'exception caught in looping job algorithm: %s' % e
            logger.warning(diagnostics)
            if "No module named" in diagnostics:
                exit_code = errors.BLACKHOLE
            else:
                exit_code = errors.UNKNOWNEXCEPTION
            return exit_code, diagnostics
        else:
            if exit_code != 0:
                return exit_code, diagnostics

        # update the ct_proxy with the current time
        mt.update('ct_looping')

    return 0, ""


def verify_disk_usage(current_time, mt, job):
    """
    Verify the disk usage.
    The function checks 1) payload stdout size, 2) local space, 3) work directory size, 4) output file sizes.

    :param current_time: current time at the start of the monitoring loop (int).
    :param mt: measured time object.
    :param job: job object.
    :return: exit code (int), error diagnostics (string).
    """

    disk_space_verification_time = convert_to_int(config.Pilot.disk_space_verification_time, default=300)
    if current_time - mt.get('ct_diskspace') > disk_space_verification_time:
        # time to check the disk space

        # check the size of the payload stdout
        exit_code, diagnostics = check_payload_stdout(job)
        if exit_code != 0:
            return exit_code, diagnostics

        # check the local space, if it's enough left to keep running the job
        exit_code, diagnostics = check_local_space()
        if exit_code != 0:
            return exit_code, diagnostics

        # check the size of the workdir
        exit_code, diagnostics = check_work_dir(job)
        if exit_code != 0:
            return exit_code, diagnostics

        # check the output file sizes
        exit_code, diagnostics = check_output_file_sizes(job)
        if exit_code != 0:
            return exit_code, diagnostics

        # update the ct_diskspace with the current time
        mt.update('ct_diskspace')

    return 0, ""


def verify_running_processes(current_time, mt, pid):
    """
    Verify the number of running processes.
    The function sets the environmental variable PILOT_MAXNPROC to the maximum number of found (child) processes
    corresponding to the main payload process id.
    The function does not return an error code (always returns exit code 0).

    :param current_time: current time at the start of the monitoring loop (int).
    :param mt: measured time object.
    :param pid: payload process id (int).
    :return: exit code (int), error diagnostics (string).
    """

    nproc_env = 0

    process_verification_time = convert_to_int(config.Pilot.process_verification_time, default=300)
    if current_time - mt.get('ct_process') > process_verification_time:
        # time to check the number of processes
        nproc = get_number_of_child_processes(pid)
        try:
            nproc_env = int(os.environ.get('PILOT_MAXNPROC', 0))
        except Exception as e:
            logger.warning('failed to convert PILOT_MAXNPROC to int: %s' % e)
        else:
            if nproc > nproc_env:
                # set the maximum number of found processes
                os.environ['PILOT_MAXNPROC'] = str(nproc)

        if nproc_env > 0:
            logger.info('maximum number of monitored processes: %d' % nproc_env)

    return 0, ""


def utility_monitor(job):
    """
    Make sure that any utility commands are still running.
    In case a utility tool has crashed, this function may restart the process.
    The function is used by the job monitor thread.

    :param job: job object.
    :return:
    """

    pilot_user = os.environ.get('PILOT_USER', 'generic').lower()
    usercommon = __import__('pilot.user.%s.common' % pilot_user, globals(), locals(), [pilot_user], 0)  # Python 2/3

    # loop over all utilities
    for utcmd in list(job.utilities.keys()):  # E.g. utcmd = MemoryMonitor, Python 2/3

        # make sure the subprocess is still running
        utproc = job.utilities[utcmd][0]
        if not utproc.poll() is None:
            if job.state == 'finished' or job.state == 'failed' or job.state == 'stageout':
                logger.debug('no need to restart utility command since payload has finished running')
                continue

            # if poll() returns anything but None it means that the subprocess has ended - which it
            # should not have done by itself
            utility_subprocess_launches = job.utilities[utcmd][1]
            if utility_subprocess_launches <= 5:
                logger.warning('detected crashed utility subprocess - will restart it')
                utility_command = job.utilities[utcmd][2]

                try:
                    proc1 = execute(utility_command, workdir=job.workdir, returnproc=True, usecontainer=False,
                                    stdout=PIPE, stderr=PIPE, cwd=job.workdir, queuedata=job.infosys.queuedata)
                except Exception as e:
                    logger.error('could not execute: %s' % e)
                else:
                    # store process handle in job object, and keep track on how many times the
                    # command has been launched
                    job.utilities[utcmd] = [proc1, utility_subprocess_launches + 1, utility_command]
            else:
                logger.warning('detected crashed utility subprocess - too many restarts, will not restart %s again' % utcmd)
        else:  # check the utility output (the selector option adds a substring to the output file name)
            filename = usercommon.get_utility_command_output_filename(utcmd, selector=True)
            path = os.path.join(job.workdir, filename)
            if os.path.exists(path):
                logger.info('file: %s exists' % path)
            else:
                logger.warning('file: %s does not exist' % path)

            # rest
            time.sleep(10)


def get_local_size_limit_stdout(bytes=True):
    """
    Return a proper value for the local size limit for payload stdout (from config file).

    :param bytes: boolean (if True, convert kB to Bytes).
    :return: size limit (int).
    """

    try:
        localsizelimit_stdout = int(config.Pilot.local_size_limit_stdout)
    except Exception as e:
        localsizelimit_stdout = 2097152
        logger.warning('bad value in config for local_size_limit_stdout: %s (will use value: %d kB)' %
                       (e, localsizelimit_stdout))

    # convert from kB to B
    if bytes:
        localsizelimit_stdout *= 1024

    return localsizelimit_stdout


def check_payload_stdout(job):
    """
    Check the size of the payload stdout.

    :param job: job object.
    :return: exit code (int), diagnostics (string).
    """

    exit_code = 0
    diagnostics = ""

    # get list of log files
    file_list = glob(os.path.join(job.workdir, 'log.*'))

    # is this a multi-trf job?
    n_jobs = job.jobparams.count("\n") + 1
    for _i in range(n_jobs):
        # get name of payload stdout file created by the pilot
        _stdout = config.Payload.payloadstdout
        if n_jobs > 1:
            _stdout = _stdout.replace(".txt", "_%d.txt" % (_i + 1))

        # add the primary stdout file to the fileList
        file_list.append(os.path.join(job.workdir, _stdout))

    # now loop over all files and check each individually (any large enough file will fail the job)
    for filename in file_list:

        if "job.log.tgz" in filename:
            logger.info("skipping file size check of file (%s) since it is a special log file" % (filename))
            continue

        if os.path.exists(filename):
            try:
                # get file size in bytes
                fsize = os.path.getsize(filename)
            except Exception as e:
                logger.warning("could not read file size of %s: %s" % (filename, e))
            else:
                # is the file too big?
                localsizelimit_stdout = get_local_size_limit_stdout()
                if fsize > localsizelimit_stdout:
                    exit_code = errors.STDOUTTOOBIG
                    diagnostics = "Payload stdout file too big: %d B (larger than limit %d B)" % \
                                  (fsize, localsizelimit_stdout)
                    logger.warning(diagnostics)

                    # kill the job
                    set_pilot_state(job=job, state="failed")
                    job.piloterrorcodes, job.piloterrordiags = errors.add_error_code(exit_code)
                    kill_processes(job.pid)

                    # remove the payload stdout file after the log extracts have been created

                    # remove any lingering input files from the work dir
                    lfns, guids = job.get_lfns_and_guids()
                    if lfns:
                        # remove any lingering input files from the work dir
                        exit_code = remove_files(job.workdir, lfns)
                else:
                    logger.info("payload stdout (%s) within allowed size limit (%d B): %d B" % (_stdout, localsizelimit_stdout, fsize))
        else:
            logger.info("skipping file size check of payload stdout file (%s) since it has not been created yet" % _stdout)

    return exit_code, diagnostics


def check_local_space():
    """
    Do we have enough local disk space left to run the job?

    :return: pilot error code (0 if success, NOLOCALSPACE if failure)
    """

    ec = 0
    diagnostics = ""

    # is there enough local space to run a job?
    cwd = os.getcwd()
    logger.debug('checking local space on %s' % cwd)
    spaceleft = convert_mb_to_b(get_local_disk_space(cwd))  # B (diskspace is in MB)
    free_space_limit = human2bytes(config.Pilot.free_space_limit)
    if spaceleft <= free_space_limit:
        diagnostics = 'too little space left on local disk to run job: %d B (need > %d B)' %\
                      (spaceleft, free_space_limit)
        ec = errors.NOLOCALSPACE
        logger.warning(diagnostics)
    else:
        logger.info('sufficient remaining disk space (%d B)' % spaceleft)

    return ec, diagnostics


def check_work_dir(job):
    """
    Check the size of the work directory.
    The function also updates the workdirsizes list in the job object.

    :param job: job object.
    :return: exit code (int), error diagnostics (string)
    """

    exit_code = 0
    diagnostics = ""

    if os.path.exists(job.workdir):
        # get the limit of the workdir
        maxwdirsize = get_max_allowed_work_dir_size(job.infosys.queuedata)

        if os.path.exists(job.workdir):
            workdirsize = get_directory_size(directory=job.workdir)

            # is user dir within allowed size limit?
            if workdirsize > maxwdirsize:
                exit_code = errors.USERDIRTOOLARGE
                diagnostics = "work directory (%s) is too large: %d B (must be < %d B)" % \
                              (job.workdir, workdirsize, maxwdirsize)
                logger.fatal("%s" % diagnostics)

                cmd = 'ls -altrR %s' % job.workdir
                _ec, stdout, stderr = execute(cmd, mute=True)
                logger.info("%s: %s" % (cmd + '\n', stdout))

                # kill the job
                # pUtil.createLockFile(True, self.__env['jobDic'][k][1].workdir, lockfile="JOBWILLBEKILLED")
                set_pilot_state(job=job, state="failed")
                job.piloterrorcodes, job.piloterrordiags = errors.add_error_code(exit_code)
                kill_processes(job.pid)

                # remove any lingering input files from the work dir
                lfns, guids = job.get_lfns_and_guids()
                if lfns:
                    remove_files(job.workdir, lfns)

                    # remeasure the size of the workdir at this point since the value is stored below
                    workdirsize = get_directory_size(directory=job.workdir)
            else:
                logger.info("size of work directory %s: %d B (within %d B limit)" % (job.workdir, workdirsize, maxwdirsize))

            # Store the measured disk space (the max value will later be sent with the job metrics)
            if workdirsize > 0:
                job.add_workdir_size(workdirsize)
        else:
            logger.warning('job work dir does not exist: %s' % job.workdir)
    else:
        logger.warning('skipping size check of workdir since it has not been created yet')

    return exit_code, diagnostics


def get_max_allowed_work_dir_size(queuedata):
    """
    Return the maximum allowed size of the work directory.

    :param queuedata: job.infosys.queuedata object.
    :return: max allowed work dir size in Bytes (int).
    """

    try:
        maxwdirsize = convert_mb_to_b(get_maximum_input_sizes())  # from MB to B, e.g. 16336 MB -> 17,129,537,536 B
    except Exception as e:
        max_input_size = get_max_input_size()
        maxwdirsize = max_input_size + config.Pilot.local_size_limit_stdout * 1024
        logger.info("work directory size check will use %d B as a max limit (maxinputsize [%d B] + local size limit for"
                    " stdout [%d B])" % (maxwdirsize, max_input_size, config.Pilot.local_size_limit_stdout * 1024))
        logger.warning('conversion caught exception: %s' % e)
    else:
        # grace margin, as discussed in https://its.cern.ch/jira/browse/ATLASPANDA-482
        margin = 10.0  # percent, read later from somewhere
        maxwdirsize = int(maxwdirsize * (1 + margin / 100.0))
        logger.info("work directory size check will use %d B as a max limit (10%% grace limit added)" % maxwdirsize)

    return maxwdirsize


def get_max_input_size(queuedata, megabyte=False):
    """
    Return a proper maxinputsize value.

    :param queuedata: job.infosys.queuedata object.
    :param megabyte: return results in MB (Boolean).
    :return: max input size (int).
    """

    _maxinputsize = queuedata.maxwdir  # normally 14336+2000 MB
    max_input_file_sizes = 14 * 1024 * 1024 * 1024  # 14 GB, 14336 MB (pilot default)
    max_input_file_sizes_mb = 14 * 1024  # 14336 MB (pilot default)
    if _maxinputsize != "":
        try:
            if megabyte:  # convert to MB int
                _maxinputsize = int(_maxinputsize)  # MB
            else:  # convert to B int
                _maxinputsize = int(_maxinputsize) * 1024 * 1024  # MB -> B
        except Exception as e:
            logger.warning("schedconfig.maxinputsize: %s" % e)
            if megabyte:
                _maxinputsize = max_input_file_sizes_mb
            else:
                _maxinputsize = max_input_file_sizes
    else:
        if megabyte:
            _maxinputsize = max_input_file_sizes_mb
        else:
            _maxinputsize = max_input_file_sizes

    if megabyte:
        logger.info("max input size = %d MB (pilot default)" % _maxinputsize)
    else:
        logger.info("Max input size = %d B (pilot default)" % _maxinputsize)

    return _maxinputsize


def check_output_file_sizes(job):
    """
    Are the output files within the allowed size limits?

    :param job: job object.
    :return: exit code (int), error diagnostics (string)
    """

    exit_code = 0
    diagnostics = ""

    # loop over all known output files
    for fspec in job.outdata:
        path = os.path.join(job.workdir, fspec.lfn)
        if os.path.exists(path):
            # get the current file size
            fsize = get_local_file_size(path)
            max_fsize = human2bytes(config.Pilot.maximum_output_file_size)
            if fsize and fsize < max_fsize:
                logger.info('output file %s is within allowed size limit (%d B < %d B)' % (path, fsize, max_fsize))
            else:
                exit_code = errors.OUTPUTFILETOOLARGE
                diagnostics = 'output file %s is not within allowed size limit (%d B > %d B)' % (path, fsize, max_fsize)
                logger.warning(diagnostics)
        else:
            logger.info('output file size check: skipping output file %s since it does not exist' % path)

    return exit_code, diagnostics



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Paul Nilsson, paul.nilsson@cern.ch, 2019

import os
import re
import tarfile

from pilot.common.exception import FileHandlingFailure, PilotException
from pilot.util.filehandling import write_file, mkdirs, rmdirs

import logging
logger = logging.getLogger(__name__)


def extract_version(name):
    """
    Try to extract the version from the DBRelease string.

    :param name: DBRelease (string).
    :return: version (string).
    """
    version = ""

    re_v = re.compile(r'DBRelease-(\d+\.\d+\.\d+)\.tar\.gz')  # Python 3 (added r)
    v = re_v.search(name)
    if v:
        version = v.group(1)
    else:
        re_v = re.compile(r'DBRelease-(\d+\.\d+\.\d+\.\d+)\.tar\.gz')  # Python 3 (added r)
        v = re_v.search(name)
        if v:
            version = v.group(1)

    return version


def get_dbrelease_version(jobpars):
    """
    Get the DBRelease version from the job parameters.

    :param jobpars: job parameters (string).
    :return: DBRelease version (string).
    """

    return extract_version(jobpars)


def get_dbrelease_dir():
    """
    Return the proper DBRelease directory

    :return: path to DBRelease (string).
    """

    path = os.path.expandvars('$VO_ATLAS_SW_DIR/database/DBRelease') if 'VO_ATLAS_SW_DIR' in os.environ else os.path.expandvars('$OSG_APP/database/DBRelease')
    if path == "" or path.startswith('OSG_APP'):
        logger.warning("note: the DBRelease database directory is not available (will not attempt to skip DBRelease stage-in)")
    else:
        if os.path.exists(path):
            logger.info("local DBRelease path verified: %s (will attempt to skip DBRelease stage-in)" % path)
        else:
            logger.warning("note: local DBRelease path does not exist: %s (will not attempt to skip DBRelease stage-in)" % path)

    return path


def is_dbrelease_available(version):
    """
    Check whether a given DBRelease file is already available.

    :param version: DBRelease version (string).
    :return: Boolean (True is DBRelease is locally available).
    """

    status = False

    # do not proceed if
    if 'ATLAS_DBREL_DWNLD' in os.environ:
        logger.info("ATLAS_DBREL_DWNLD is set: do not skip DBRelease stage-in")
        return status

    # get the local path to the DBRelease directory
    path = get_dbrelease_dir()

    if path != "" and os.path.exists(path):
        # get the list of available DBRelease directories
        dir_list = os.listdir(path)

        # is the required DBRelease version available?
        if dir_list:
            if version in dir_list:
                logger.info("found version %s in path %s (%d releases found)" % (version, path, len(dir_list)))
                status = True
            else:
                logger.warning("did not find version %s in path %s (%d releases found)" % (version, path, len(dir_list)))
        else:
            logger.warning("empty DBRelease directory list: %s" % path)
    else:
        logger.warning('no such DBRelease path: %s' % path)

    return status


def create_setup_file(version, path):
    """
    Create the DBRelease setup file.

    :param version: DBRelease version (string).
    :param path: path to local DBReleases (string).
    :return: Boolean (True if DBRelease setup file was successfully created).
    """

    status = False

    # get the DBRelease directory
    d = get_dbrelease_dir()
    if d != "" and version != "":
        # create the python code string to be written to file
        txt = "import os\n"
        txt += "os.environ['DBRELEASE'] = '%s'\n" % version
        txt += "os.environ['DATAPATH'] = '%s/%s:' + os.environ['DATAPATH']\n" % (d, version)
        txt += "os.environ['DBRELEASE_REQUIRED'] = '%s'\n" % version
        txt += "os.environ['DBRELEASE_REQUESTED'] = '%s'\n" % version
        txt += "os.environ['CORAL_DBLOOKUP_PATH'] = '%s/%s/XMLConfig'\n" % (d, version)

        try:
            status = write_file(path, txt)
        except FileHandlingFailure as e:
            logger.warning('failed to create DBRelease setup file: %s' % e)
        else:
            logger.info("Created setup file with the following content:.................................\n%s" % txt)
            logger.info("...............................................................................")
    else:
        logger.warning('failed to create %s for DBRelease version=%s and directory=%s' % (path, version, d))

    return status


def create_dbrelease(version, path):
    """
    Create the DBRelease file only containing a setup file.

    :param version: DBRelease version (string).
    :param path: path to DBRelease (string).
    :return: Boolean (True is DBRelease file was successfully created).
    """

    status = False

    # create the DBRelease and version directories
    dbrelease_path = os.path.join(path, 'DBRelease')
    _path = os.path.join(dbrelease_path, version)
    try:
        mkdirs(_path, chmod=None)
    except PilotException as e:
        logger.warning('failed to create directories for DBRelease: %s' % e)
    else:
        logger.debug('created directories: %s' % _path)

        # create the setup file in the DBRelease directory
        version_path = os.path.join(dbrelease_path, version)
        setup_filename = "setup.py"
        _path = os.path.join(version_path, setup_filename)
        if create_setup_file(version, _path):
            logger.info("created DBRelease setup file: %s" % _path)

            # now create a new DBRelease tarball
            filename = os.path.join(path, "DBRelease-%s.tar.gz" % version)
            logger.info("creating file: %s" % filename)
            try:
                tar = tarfile.open(filename, "w:gz")
            except Exception as e:
                logger.warning("could not create DBRelease tar file: %s" % e)
            else:
                if tar:
                    # add the setup file to the tar file
                    tar.add("%s/DBRelease/%s/%s" % (path, version, setup_filename))

                    # create the symbolic link DBRelease/current ->  12.2.1
                    try:
                        _link = os.path.join(path, "DBRelease/current")
                        os.symlink(version, _link)
                    except Exception as e:
                        logger.warning("failed to create symbolic link %s: %s" % (_link, e))
                    else:
                        logger.warning("created symbolic link: %s" % _link)

                        # add the symbolic link to the tar file
                        tar.add(_link)

                        # done with the tar archive
                        tar.close()

                        logger.info("created new DBRelease tar file: %s" % filename)
                        status = True
                else:
                    logger.warning("failed to open DBRelease tar file")

            # clean up
            if rmdirs(dbrelease_path):
                logger.debug("cleaned up directories in path: %s" % dbrelease_path)
        else:
            logger.warning("failed to create DBRelease setup file")
            if rmdirs(dbrelease_path):
                logger.debug("cleaned up directories in path: %s" % dbrelease_path)

    return status



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Wen Guan, wen.guan@cern.ch, 2017

"""
Hooks for ARC-ControlTower EventService.
"""

from pilot.eventservice.eshook import ESHook


class ACTESHook(ESHook):
    def get_payload(self):
        """
        Get payload to execute.

        :returns: dict {'payload': <cmd string>, 'output_file': <filename or without it>, 'error_file': <filename or without it>}
        """
        raise Exception("Not Implemented")

    def get_event_ranges(self, num_ranges=1):
        """
        Get event ranges.

        :returns: dict of event ranges.
                 None if no available events.
        """
        raise Exception("Not Implemented")

    def handle_out_message(self, message):
        """
        Handle ES output or error messages.

        :param message: a dict of parsed message.
                        For 'finished' event ranges, it's {'id': <id>, 'status': 'finished', 'output': <output>, 'cpu': <cpu>,
                                                           'wall': <wall>, 'message': <full message>}.
                        Fro 'failed' event ranges, it's {'id': <id>, 'status': 'finished', 'message': <full message>}.
        """
        raise Exception("Not Implemented")



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Paul Nilsson, paul.nilsson@cern.ch, 2019

from __future__ import print_function  # Python 2, 2to3 complains about this

import functools
import signal
import threading
from time import time
from sys import stderr

try:
    import Queue as queue  # noqa: N813
except Exception:
    import queue  # Python 3

from collections import namedtuple

from pilot.common.exception import ExcThread
from pilot.control import job, data, monitor
from pilot.util.constants import SUCCESS, PILOT_KILL_SIGNAL
from pilot.util.timing import add_to_pilot_timing

import logging
logger = logging.getLogger(__name__)


def interrupt(args, signum, frame):
    """
    Interrupt function on the receiving end of kill signals.
    This function is forwarded any incoming signals (SIGINT, SIGTERM, etc) and will set abort_job which instructs
    the threads to abort the job.

    :param args: pilot arguments.
    :param signum: signal.
    :param frame: stack/execution frame pointing to the frame that was interrupted by the signal.
    :return:
    """

    try:
        sig = [v for v, k in signal.__dict__.iteritems() if k == signum][0]  # Python 2
    except Exception:
        sig = [v for v, k in list(signal.__dict__.items()) if k == signum][0]  # Python 3
    add_to_pilot_timing('0', PILOT_KILL_SIGNAL, time(), args)
    add_to_pilot_timing('1', PILOT_KILL_SIGNAL, time(), args)
    logger.warning('caught signal: %s' % sig)
    args.signal = sig
    logger.warning('will instruct threads to abort and update the server')
    args.abort_job.set()
    logger.warning('waiting for threads to finish')
    args.job_aborted.wait()
    logger.warning('setting graceful stop (in case it was not set already), pilot will abort')
    args.graceful_stop.set()


def run(args):
    """
    Main execution function for the stage-in workflow.

    The function sets up the internal queues which handle the flow of jobs.

    :param args: pilot arguments.
    :returns: traces.
    """

    logger.info('setting up signal handling')
    signal.signal(signal.SIGINT, functools.partial(interrupt, args))
    signal.signal(signal.SIGTERM, functools.partial(interrupt, args))
    signal.signal(signal.SIGQUIT, functools.partial(interrupt, args))
    signal.signal(signal.SIGSEGV, functools.partial(interrupt, args))
    signal.signal(signal.SIGXCPU, functools.partial(interrupt, args))
    signal.signal(signal.SIGUSR1, functools.partial(interrupt, args))
    signal.signal(signal.SIGBUS, functools.partial(interrupt, args))

    logger.info('setting up queues')
    queues = namedtuple('queues', ['jobs', 'data_in', 'data_out', 'current_data_in', 'validated_jobs',
                                   'finished_jobs', 'finished_data_in', 'finished_data_out',
                                   'failed_jobs', 'failed_data_in', 'failed_data_out', 'completed_jobs'])

    queues.jobs = queue.Queue()
    queues.data_in = queue.Queue()
    queues.data_out = queue.Queue()

    queues.current_data_in = queue.Queue()
    queues.validated_jobs = queue.Queue()

    queues.finished_jobs = queue.Queue()
    queues.finished_data_in = queue.Queue()
    queues.finished_data_out = queue.Queue()

    queues.failed_jobs = queue.Queue()
    queues.failed_data_in = queue.Queue()
    queues.failed_data_out = queue.Queue()

    queues.completed_jobs = queue.Queue()

    logger.info('setting up tracing')
    traces = namedtuple('traces', ['pilot'])
    traces.pilot = {'state': SUCCESS,
                    'nr_jobs': 0,
                    'error_code': 0,
                    'command': None}

    # define the threads
    targets = {'job': job.control, 'data': data.control, 'monitor': monitor.control}
    threads = [ExcThread(bucket=queue.Queue(), target=target, kwargs={'queues': queues, 'traces': traces, 'args': args},
                         name=name) for name, target in list(targets.items())]

    logger.info('starting threads')
    [thread.start() for thread in threads]

    logger.info('waiting for interrupts')

    while threading.activeCount() > 1:
        for thread in threads:
            bucket = thread.get_bucket()
            try:
                exc = bucket.get(block=False)
            except queue.Empty:
                pass
            else:
                exc_type, exc_obj, exc_trace = exc
                # deal with the exception
                print('received exception from bucket queue in generic workflow: %s' % exc_obj, file=stderr)
                # logger.fatal('caught exception: %s' % exc_obj)

            thread.join(0.1)

    logger.info('end of stage-in workflow (traces error code: %d)' % traces.pilot['error_code'])

    return traces



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Paul Nilsson, paul.nilsson@cern.ch, 2017

import os
import re
import glob
from time import sleep

from pilot.common.errorcodes import ErrorCodes
from pilot.util.container import execute
from pilot.util.filehandling import copy

import logging
logger = logging.getLogger(__name__)

errors = ErrorCodes()


def get_analysis_trf(transform, workdir):
    """
    Prepare to download the user analysis transform with curl.
    The function will verify the download location from a known list of hosts.

    :param transform: full trf path (url) (string).
    :param workdir: work directory (string).
    :return: exit code (int), diagnostics (string), transform_name (string)
    """

    ec = 0
    diagnostics = ""

    # test if $HARVESTER_WORKDIR is set
    harvester_workdir = os.environ.get('HARVESTER_WORKDIR')
    if harvester_workdir is not None:
        search_pattern = "%s/jobO.*.tar.gz" % harvester_workdir
        logger.debug("search_pattern - %s" % search_pattern)
        jobopt_files = glob.glob(search_pattern)
        for jobopt_file in jobopt_files:
            logger.debug("jobopt_file = %s workdir = %s" % (jobopt_file, workdir))
            try:
                copy(jobopt_file, workdir)
            except Exception as e:
                logger.error("could not copy file %s to %s : %s" % (jobopt_file, workdir, e))

    if '/' in transform:
        transform_name = transform.split('/')[-1]
    else:
        logger.warning('did not detect any / in %s (using full transform name)' % transform)
        transform_name = transform

    # is the command already available? (e.g. if already downloaded by a preprocess/main process step)
    if os.path.exists(os.path.join(workdir, transform_name)):
        logger.info('script %s is already available - no need to download again' % transform_name)
        return ec, diagnostics, transform_name

    original_base_url = ""

    # verify the base URL
    for base_url in get_valid_base_urls():
        if transform.startswith(base_url):
            original_base_url = base_url
            break

    if original_base_url == "":
        diagnostics = "invalid base URL: %s" % transform
        return errors.TRFDOWNLOADFAILURE, diagnostics, ""

    # try to download from the required location, if not - switch to backup
    status = False
    for base_url in get_valid_base_urls(order=original_base_url):
        trf = re.sub(original_base_url, base_url, transform)
        logger.debug("attempting to download script: %s" % trf)
        status, diagnostics = download_transform(trf, transform_name, workdir)
        if status:
            break

    if not status:
        return errors.TRFDOWNLOADFAILURE, diagnostics, ""

    logger.info("successfully downloaded script")
    path = os.path.join(workdir, transform_name)
    logger.debug("changing permission of %s to 0o755" % path)
    try:
        os.chmod(path, 0o755)  # Python 2/3
    except Exception as e:
        diagnostics = "failed to chmod %s: %s" % (transform_name, e)
        return errors.CHMODTRF, diagnostics, ""

    return ec, diagnostics, transform_name


def get_valid_base_urls(order=None):
    """
    Return a list of valid base URLs from where the user analysis transform may be downloaded from.
    If order is defined, return given item first.
    E.g. order=http://atlpan.web.cern.ch/atlpan -> ['http://atlpan.web.cern.ch/atlpan', ...]
    NOTE: the URL list may be out of date.

    :param order: order (string).
    :return: valid base URLs (list).
    """

    valid_base_urls = []
    _valid_base_urls = ["https://atlpan.web.cern.ch/atlpan",
                        "http://pandaserver.cern.ch"]

    if order:
        valid_base_urls.append(order)
        for url in _valid_base_urls:
            if url != order:
                valid_base_urls.append(url)
    else:
        valid_base_urls = _valid_base_urls

    return valid_base_urls


def download_transform(url, transform_name, workdir):
    """
    Download the transform from the given url
    :param url: download URL with path to transform (string).
    :param transform_name: trf name (string).
    :param workdir: work directory (string).
    :return:
    """

    status = False
    diagnostics = ""
    path = os.path.join(workdir, transform_name)
    cmd = 'curl -sS \"%s\" > %s' % (url, path)
    trial = 1
    max_trials = 3

    # test if $HARVESTER_WORKDIR is set
    harvester_workdir = os.environ.get('HARVESTER_WORKDIR')
    if harvester_workdir is not None:
        # skip curl by setting max_trials = 0
        max_trials = 0
        source_path = os.path.join(harvester_workdir, transform_name)
        try:
            copy(source_path, path)
            status = True
        except Exception as error:
            status = False
            diagnostics = "Failed to copy file %s to %s : %s" % (source_path, path, error)
            logger.error(diagnostics)

    # try to download the trf a maximum of 3 times
    while trial <= max_trials:
        logger.info("executing command [trial %d/%d]: %s" % (trial, max_trials, cmd))

        exit_code, stdout, stderr = execute(cmd, mute=True)
        if not stdout:
            stdout = "(None)"
        if exit_code != 0:
            # Analyze exit code / output
            diagnostics = "curl command failed: %d, %s, %s" % (exit_code, stdout, stderr)
            logger.warning(diagnostics)
            if trial == max_trials:
                logger.fatal('could not download transform: %s' % stdout)
                status = False
                break
            else:
                logger.info("will try again after 60 s")
                sleep(60)
        else:
            logger.info("curl command returned: %s" % stdout)
            status = True
            break
        trial += 1

    return status, diagnostics



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Wen Guan, wen.guan@cern.ch, 2018


import logging
logger = logging.getLogger(__name__)

"""
A factory to manage plugins
"""


class PluginFactory(object):

    def __init__(self, *args, **kwargs):
        self.classMap = {}

    def get_plugin(self, confs):
        """
        Load plugin class

        :param confs: a dict of configurations.
        """

        class_name = confs['class']
        if class_name is None:
            logger.error("[class] is not defined in confs: %s" % confs)
            return None

        if class_name not in self.classMap:
            logger.info("Trying to import %s" % class_name)
            components = class_name.split('.')
            mod = __import__('.'.join(components[:-1]))
            for comp in components[1:]:
                mod = getattr(mod, comp)
            self.classMap[class_name] = mod

        args = {}
        for key in confs:
            if key in ['class']:
                continue
            args[key] = confs[key]

        cls = self.classMap[class_name]
        logger.info("Importing %s with args: %s" % (cls, args))
        impl = cls(**args)

        return impl



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Wen Guan, wen.guan@cern.ch, 2018

"""
Main classes to manage the messages between ES and harvester/ACT/Panda.
"""

import json
import logging
import os
import threading
import time
try:
    import Queue as queue  # noqa: N813
except Exception:
    import queue  # Python 3

from pilot.common import exception
from pilot.common.pluginfactory import PluginFactory


logger = logging.getLogger(__name__)


"""
Communication response
"""


class CommunicationResponse(object):
    def __init__(self, attrs=None):
        if not attrs:
            attrs = {}
        if not isinstance(attrs, dict):
            attrs = json.loads(attrs)

        def_attrs = {'status': None, 'content': None, 'exception': None}

        for key in def_attrs:
            if key not in attrs:
                attrs[key] = def_attrs[key]

        for key in attrs:
            setattr(self, key, attrs[key])

    def __str__(self):
        json_str = {}
        for key, value in list(self.__dict__.items()):  # Python 2/3
            if value and type(value) is list:
                json_str[key] = []
                for list_item in value:
                    json_str[key].append(str(list_item))
            elif value:
                json_str[key] = str(value)
            else:
                json_str[key] = value
        return json.dumps(json_str)


"""
Communication request
"""


class CommunicationRequest(object):
    class RequestType(object):
        RequestJobs = 'request_jobs'
        UpdateJobs = 'update_jobs'
        RequestEvents = 'request_events'
        UpdateEvents = 'update_events'

    def __init__(self, attrs=None):
        if not attrs:
            attrs = {}
        if not isinstance(attrs, dict):
            attrs = json.loads(attrs)

        if attrs['request_type'] == CommunicationRequest.RequestType.RequestJobs:
            def_attrs = {'num_jobs': 1, 'post_hook': None, 'response': None}
        if attrs['request_type'] == CommunicationRequest.RequestType.RequestEvents:
            def_attrs = {'num_event_ranges': 1, 'post_hook': None, 'response': None}
        if attrs['request_type'] == CommunicationRequest.RequestType.UpdateEvents:
            def_attrs = {'update_events': None, 'post_hook': None, 'response': None}
        if attrs['request_type'] == CommunicationRequest.RequestType.UpdateJobs:
            def_attrs = {'jobs': None, 'post_hook': None, 'response': None}

        for key in def_attrs:
            if key not in attrs:
                attrs[key] = def_attrs[key]

        for key in attrs:
            setattr(self, key, attrs[key])
        self.abort = False

    def __str__(self):
        json_str = {}
        for key, value in list(self.__dict__.items()):  # Python 2/3
            if value and type(value) is list:
                json_str[key] = []
                for list_item in value:
                    json_str[key].append(str(list_item))
            elif value:
                json_str[key] = str(value)
            else:
                json_str[key] = value
        return json.dumps(json_str)


"""
Communication manager thread
"""


class CommunicationManager(threading.Thread, PluginFactory):

    def __init__(self, *args, **kwargs):
        super(CommunicationManager, self).__init__()
        PluginFactory.__init__(self, *args, **kwargs)
        self.setName("CommunicationManager")
        self.post_get_jobs = None
        self.post_get_event_ranges_hook = None
        self.queues = {'request_get_jobs': queue.Queue(),  # Python 2/3
                       'update_jobs': queue.Queue(),
                       'request_get_events': queue.Queue(),
                       'update_events': queue.Queue(),
                       'processing_get_jobs': queue.Queue(),
                       'processing_update_jobs': queue.Queue(),
                       'processing_get_events': queue.Queue(),
                       'processing_update_events': queue.Queue()}
        self.queue_limits = {'request_get_jobs': None,
                             'update_jobs': None,
                             'request_get_events': None,
                             'update_events': None,
                             'processing_get_jobs': 1,
                             'processing_update_jobs': 1,
                             'processing_get_events': 1,
                             'processing_update_events': 1}
        self.stop_event = threading.Event()
        self.args = args
        self.kwargs = kwargs

    def stop(self):
        """
        Set stop signal(main run process will clean queued requests to release waiting clients and then quit)
        """
        if not self.is_stop():
            logger.info("Stopping Communication Manager.")
            self.stop_event.set()

    def is_stop(self):
        """
        check whether the stop signal is set

        :returns: True if the stop signal is set, otherwise False
        """
        return self.stop_event.is_set()

    def get_jobs(self, njobs=1, post_hook=None, args=None):
        """
        Function can be called by client to send a get_job request and get a response with jobs.

        :returns: jobs(got from jobs servers)
        :raise: Exception catched when getting jobs
        """

        if self.is_stop():
            return None

        req_attrs = {}
        if args:
            if not type(args) is dict:
                args = vars(args)
            for key, value in list(args.items()):  # Python 2/3
                req_attrs[key] = value

        other_req_attrs = {'request_type': CommunicationRequest.RequestType.RequestJobs,
                           'num_jobs': njobs,
                           'post_hook': post_hook}
        for key, value in list(other_req_attrs.items()):  # Python 2/3
            req_attrs[key] = value

        req = CommunicationRequest(req_attrs)
        self.queues['request_get_jobs'].put(req)

        if req.post_hook:
            return

        while req.response is None:
            time.sleep(1)
        if req.response.exception:
            raise req.response.exception
        if req.response.status is False:
            return None
        else:
            return req.response.content

    def update_jobs(self, jobs, post_hook=None):
        """
        Function can be called by client to update jobs' status to server.

        :returns: status of updating jobs
        :raise: Exception catched when updating jobs
        """

        if self.is_stop():
            return None

        req_attrs = {'request_type': CommunicationRequest.RequestType.UpdateJobs,
                     'jobs': jobs,
                     'post_hook': post_hook}

        req = CommunicationRequest(req_attrs)
        self.queues['update_jobs'].put(req)

        if req.post_hook:
            return

        while req.response is None:
            time.sleep(1)
        if req.response.exception:
            raise req.response.exception
        if req.response.status is False:
            return None
        else:
            return req.response.content

    def get_event_ranges(self, num_event_ranges=1, post_hook=None, job=None):
        """
        Function can be called by client to send a get_event_ranges request and get a response with event ranges.

        :returns: event ranges (got from jobs servers)
        :raise: Exception caught when getting event ranges
        """

        if self.is_stop():
            return None

        if not job:
            resp_attrs = {'status': -1,
                          'content': None,
                          'exception': exception.CommunicationFailure("Get events failed because job info missing(job: %s)" % job)}
            resp = CommunicationResponse(resp_attrs)
            raise resp.exception

        req_attrs = {'request_type': CommunicationRequest.RequestType.RequestEvents,
                     'num_event_ranges': num_event_ranges,
                     'post_hook': post_hook}
        req_attrs['jobid'] = job['PandaID']
        req_attrs['jobsetid'] = job['jobsetID']
        req_attrs['taskid'] = job['taskID']
        req_attrs['num_ranges'] = num_event_ranges

        req = CommunicationRequest(req_attrs)
        self.queues['request_get_events'].put(req)

        if req.post_hook:
            return

        while req.response is None:
            time.sleep(1)
        if req.response.exception:
            raise req.response.exception
        if req.response.status is False:
            return None
        else:
            return req.response.content

    def update_events(self, update_events, post_hook=None):
        """
        Function can be called by client to send a update_events request.

        :returns: status of updating event ranges
        :raise: Exception catched when updating event ranges
        """

        if self.is_stop():
            return None

        req_attrs = {'request_type': CommunicationRequest.RequestType.UpdateEvents,
                     'update_events': update_events,
                     'post_hook': post_hook}
        req = CommunicationRequest(req_attrs)
        self.queues['update_events'].put(req)

        if req.post_hook:
            return

        while req.response is None:
            time.sleep(1)
        if req.response.exception:
            raise req.response.exception
        if req.response.status is False:
            return None
        else:
            return req.response.content

    def get_plugin_confs(self):
        """
        Get different plugin for different communicator

        :returns: dict with {'class': <plugin_class>} and other items
        """

        plugin = os.environ.get('COMMUNICATOR_PLUGIN', None)
        if not plugin:
            plugin_confs = {'class': 'pilot.eventservice.communicationmanager.plugins.pandacommunicator.PandaCommunicator'}
        elif plugin == 'act':
            plugin_confs = {'class': 'pilot.eventservice.communicationmanager.plugins.actcommunicator.ACTCommunicator'}
        elif plugin == 'harvestersf':
            plugin_confs = {'class': 'pilot.eventservice.communicationmanager.plugins.harvestersharefilecommunicator.HarvesterShareFileCommunicator'}
        else:
            plugin_confs = {'class': 'pilot.eventservice.communicationmanager.plugins.pandacommunicator.PandaCommunicator'}

        if self.args:
            for key, value in list(vars(self.args).items()):  # Python 2/3
                plugin_confs[key] = value
        return plugin_confs

    def can_process_request(self, processor, process_type):
        """
        To check whether it is ready to process request in a type.
        For request such as HarvesterShareFileCommunicator, it should check whether there are processing requests to avoid overwriting files.

        :returns: True or False
        """

        if self.queues[process_type].empty():
            return False

        next_queue = processor[process_type]['next_queue']
        if next_queue is None or self.queue_limits[next_queue] is None:
            return True

        if self.queues[next_queue].qsize() < self.queue_limits[next_queue]:
            return True

        return False

    def run(self):
        """
        Main loop to handle communication requests
        """

        confs = self.get_plugin_confs()
        logger.info("Communication plugin confs: %s" % confs)
        communicator = self.get_plugin(confs)
        logger.info("Communication: %s" % communicator)

        processor = {'request_get_jobs': {'pre_check': communicator.pre_check_get_jobs,
                                          'handler': communicator.request_get_jobs,
                                          'next_queue': 'processing_get_jobs',
                                          'process_req_post_hook': False},
                     'request_get_events': {'pre_check': communicator.pre_check_get_events,
                                            'handler': communicator.request_get_events,
                                            'next_queue': 'processing_get_events',
                                            'process_req_post_hook': False},
                     'update_jobs': {'pre_check': communicator.pre_check_update_jobs,
                                     'handler': communicator.update_jobs,
                                     'next_queue': None,
                                     'process_req_post_hook': True},
                     'update_events': {'pre_check': communicator.pre_check_update_events,
                                       'handler': communicator.update_events,
                                       'next_queue': None,
                                       'process_req_post_hook': True},
                     'processing_get_jobs': {'pre_check': communicator.check_get_jobs_status,
                                             'handler': communicator.get_jobs,
                                             'next_queue': None,
                                             'process_req_post_hook': True},
                     'processing_get_events': {'pre_check': communicator.check_get_events_status,
                                               'handler': communicator.get_events,
                                               'next_queue': None,
                                               'process_req_post_hook': True}
                     }

        while True:
            has_req = False
            for process_type in processor:
                if self.is_stop():
                    while not self.queues[process_type].empty():
                        req = self.queues[process_type].get()
                        logger.info("Is going to stop, aborting request: %s" % req)
                        req.abort = True
                        resp_attrs = {'status': None,
                                      'content': None,
                                      'exception': exception.CommunicationFailure("Communication manager is stopping, abort this request")}
                        req.response = CommunicationResponse(resp_attrs)
                elif self.can_process_request(processor, process_type):
                    pre_check_resp = processor[process_type]['pre_check']()
                    if not pre_check_resp.status == 0:
                        continue

                    logger.info("Processing %s" % process_type)

                    has_req = True
                    req = self.queues[process_type].get()

                    logger.info("Processing %s request: %s" % (process_type, req))
                    res = processor[process_type]['handler'](req)
                    logger.info("Processing %s respone: %s" % (process_type, res))

                    if res.status is False:
                        req.response = res
                    else:
                        next_queue = processor[process_type]['next_queue']
                        if next_queue:
                            self.queues[next_queue].put(req)
                        else:
                            req.response = res
                        process_req_post_hook = processor[process_type]['process_req_post_hook']
                        if process_req_post_hook and req.post_hook:
                            req.post_hook(res)
            if not has_req:
                if self.is_stop():
                    break
            time.sleep(1)
        logger.info("Communication manager stopped.")



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Authors:
# - Wen Guan, wen.guan@cern.ch, 2018
# - Paul Nilsson, paul.nilsson@cern.ch, 2020

from pilot.common import exception

import logging
logger = logging.getLogger(__name__)

"""
Base communicator
"""


class BaseCommunicator(object):
    _instance = None

    def __new__(class_, *args, **kwargs):
        if not isinstance(class_._instance, class_):
            class_._instance = object.__new__(class_, *args, **kwargs)
        return class_._instance

    def __init__(self, *args, **kwargs):
        super(BaseCommunicator, self).__init__()
        for key in kwargs:
            setattr(self, key, kwargs[key])

    def pre_check_get_jobs(self, req):
        """
        Precheck whether it's ok to send a requst to get jobs.
        """
        raise exception.NotImplemented()

    def request_get_jobs(self, req):
        """
        Send a requst to get jobs.
        """
        raise exception.NotImplemented()

    def check_get_jobs_status(self, req):
        """
        Check whether jobs are prepared
        """
        raise exception.NotImplemented()

    def get_jobs(self, req):
        """
        Get the job
        """
        raise exception.NotImplemented()

    def update_jobs(self, req):
        """
        Update jobs status.
        """
        raise exception.NotImplemented()

    def pre_check_get_events(self, req):
        """
        Precheck whether it's ok to send a request to get events.
        """
        raise exception.NotImplemented()

    def request_get_events(self, req):
        """
        Send a requst to get events.
        """
        raise exception.NotImplemented()

    def check_get_events_status(self, req):
        """
        Check whether events prepared
        """
        raise exception.NotImplemented()

    def get_events(self, req):
        """
        Get events
        """
        raise exception.NotImplemented()

    def pre_check_update_events(self, req):
        """
        Precheck whether it's ok to update events.
        """
        raise exception.NotImplemented()

    def update_events(self, req):
        """
        Update events.
        """
        raise exception.NotImplemented()

    def pre_check_update_jobs(self, req):
        """
        Precheck whether it's ok to update event ranges.
        """
        raise exception.NotImplemented()
